"""Constants used by the module."""

PRIORITY_HIGH = "High"
PRIORITY_LOW = "Low"
PRIORITY_NORMAL = "Normal"

STATUS_APPROVED = "Approved"
STATUS_BLOCKED = "Blocked"
STATUS_DEPRECATED = "Deprecated"
STATUS_DONE = "Done"
STATUS_DRAFT = "Draft"
STATUS_FAIL = "Fail"
STATUS_IN_PROGRESS = "In Progress"
STATUS_NOT_EXECUTED = "Not Executed"
STATUS_PASS = "Pass"

STEP_TYPE_BY_STEP = "STEP_BY_STEP"
STEP_TYPE_PLAIN = "PLAIN_TEXT"

TEST_CASE = "TEST_CASE"
TEST_PLAN = "TEST_PLAN"
TEST_RUN = "TEST_RUN"
